/*
 * Copyright 2010 Trustees of the University of Pennsylvania Licensed under the
 * Educational Community License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License. You may
 * obtain a copy of the License at
 *
 * http://www.osedu.org/licenses/ECL-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS"
 * BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */
package metridoc.workflows

import groovy.util.logging.Slf4j
import java.sql.BatchUpdateException
import java.sql.SQLException
import java.util.concurrent.TimeUnit
import java.util.concurrent.atomic.AtomicInteger
import javax.sql.DataSource
import metridoc.camel.builder.GroovyRouteBuilder
import metridoc.camel.component.poll.PollEndpoint
import metridoc.dsl.JobBuilder
import metridoc.plugins.impl.iterators.IteratorFactory
import metridoc.plugins.iterators.IteratorCreator
import metridoc.plugins.sql.SqlPlus
import metridoc.utils.SystemUtils
import metridoc.workflows.validation.impl.FileIngestorValidationErrorHandler
import org.apache.camel.CamelContext
import org.apache.camel.Endpoint
import org.apache.camel.Exchange
import org.apache.camel.builder.RouteBuilder
import org.apache.camel.component.file.GenericFile
import org.apache.camel.component.file.GenericFileEndpoint
import org.apache.camel.component.file.GenericFileFilter
import org.apache.camel.model.AggregateDefinition
import org.apache.camel.model.ProcessorDefinition

/**
 * Created by IntelliJ IDEA.
 * User: tbarker
 * Date: 2/21/12
 * Time: 3:26 PM
 */
@Slf4j
class FileToLoadingTableWorkflow extends Workflow {

    public static final String DEFAULT_WORKING_DIRECTORY = "${SystemUtils.METRIDOC_HOME}${SystemUtils.FILE_SEPARATOR}files"
    CamelContext camelContext
    IteratorCreator iteratorCreator
    String workingDirectory
    String fileNameFilter
    Closure fileFilter
    /**
     * the uri used as a camel endpoint
     */
    String workingDirectoryUri
    int batchSize = 5000
    boolean parallel = true
    int maxNumberOfFiles = 1
    String loadingTable
    String loadingTableSql
    DataSource dataSource
    SqlPlus sqlPlus
    String projectName
    Integer previewLines = Integer.MAX_VALUE
    int previewOffset = 0
    long waitForLocalFilePoll = 3000L
    int updateAt = 10000
    int currentEndOfLine
    boolean deleteLoadingFileWhenDone = true
    String moveLoadingFileWhenDone
    String moveLoadingFileOnError
    Endpoint endpoint
    boolean noop = false
    String validationErrorTable = FileIngestorValidationErrorHandler.DEFAULT_VALIDATION_TABLE

    SqlPlus getSqlPlus() {
        if (sqlPlus) {
            return sqlPlus
        }

        sqlPlus = new SqlPlus(dataSource)
    }

    String getLoadingTable() {
        if (loadingTable) {
            return loadingTable
        }

        loadingTable = "${projectName}_loading"
    }

    String getLoadingTableSql() {
        if (loadingTableSql) {
            return loadingTableSql
        }

        loadingTableSql = getLoadingTable()
    }

    Closure aggregateRecords = {ProcessorDefinition definition ->
        def aggDef = definition.aggregateBody(batchSize) as AggregateDefinition
        return aggDef.forceCompletionOnStop().process {
            CurrentProgressLogger.instance.update("parse, validate and aggregate lines", it.in.body.size(), updateAt)
        }
    }

    Closure splitContents = {ProcessorDefinition definition, GroovyRouteBuilder builder ->

        def currentDefinition = definition.process {Exchange exchange ->
            def iterator = getIteratorCreator().create(exchange.in.body, exchange.in.headers)
            exchange.in.body = new IteratorWrapper(
                iterator: iterator,
                limit: previewLines,
                offset: previewOffset
            )
        }
            .split(builder.body()).streaming()
        if (parallel) {
            currentDefinition = currentDefinition.parallelProcessing()
        }

        def loggedWarning = false
        currentDefinition = currentDefinition.filter {
            if (builder.routeException) {
                if (!loggedWarning) {
                    log.warn("splitter is ignoring the rest of the messages since there was an exception")
                    loggedWarning = true
                }
                return false
            }
            return true
        }

        return currentDefinition
    }

    Closure loadData = {ProcessorDefinition definition ->
        definition.process {Exchange exchange ->
            def records = exchange.in.body

            def sqlPlus = getSqlPlus()
            try {
                sqlPlus.runBatch(loadingTableSql, records)
            } catch (BatchUpdateException ex) {
                log.warn("exception occurred while trying to insert a batch, will process the records line by line to find the culprit")

                if (records instanceof List) {
                    Map record
                    try {
                        records.each {
                            record = it
                            sqlPlus.runBatch(loadingTableSql, record)
                        }
                    } catch (SQLException sqlException) {
                        log.error("could not insert record with line {} from file {} because {}", record.line_num + previewOffset, record.source_file, sqlException.message)
                        throw sqlException
                    }
                } else {
                    throw ex
                }
            }
        }
    }

    Closure processLine

    Endpoint getEndpoint() {

        if (endpoint) {
            return endpoint
        }

        Endpoint wrappedEndpoint = camelContext.getEndpoint(getWorkingDirectoryUri())

        if (wrappedEndpoint instanceof GenericFileEndpoint) {
            wrappedEndpoint.filter = new FileFilter(fileFilter: fileFilter, fileNameFilter: fileNameFilter)
            wrappedEndpoint.delete = deleteLoadingFileWhenDone
            wrappedEndpoint.move = moveLoadingFileWhenDone
            wrappedEndpoint.moveFailed = moveLoadingFileOnError
            wrappedEndpoint.noop = noop
        }

        endpoint = new PollEndpoint(wrappedEndpoint: wrappedEndpoint, maxMessages: maxNumberOfFiles, camelContext: camelContext, wait: waitForLocalFilePoll)
    }

    String getWorkingDirectory() {
        if (workingDirectory) {
            return workingDirectory
        }

        def metridocHome = SystemUtils.METRIDOC_HOME
        def slash = SystemUtils.FILE_SEPARATOR

        if (projectName) {
            workingDirectory = "${metridocHome}${slash}${projectName}${slash}files"
        } else {
            workingDirectory = DEFAULT_WORKING_DIRECTORY
        }

        return workingDirectory
    }

    String getWorkingDirectoryUri() {
        if (workingDirectoryUri) {
            return workingDirectoryUri
        }
        workingDirectoryUri = "file://${getWorkingDirectory()}"
    }

    IteratorCreator getIteratorCreator() {
        if (iteratorCreator != null) {
            return iteratorCreator
        }

        iteratorCreator = new IteratorFactory()
    }

    @Override
    Object run() {
        JobBuilder.isJob(this)

        def fileList = []

        AtomicInteger lineCount = new AtomicInteger(0)
        runRoute(camelContext: camelContext) {
            ProcessorDefinition currentDefinition = from(getEndpoint()).process {
                def fileName = it.in.getHeader(Exchange.FILE_NAME_ONLY, String.class)
                log.info "processing file {}", fileName
                fileList.add(fileName)
            }

            currentDefinition = runClosure(splitContents, currentDefinition, delegate)

            currentDefinition.process {Exchange exchange ->
                lineCount.getAndIncrement()
            }

            currentDefinition = runClosure(processLine, currentDefinition, delegate)
            currentDefinition = runClosure(aggregateRecords, currentDefinition, delegate)

            runClosure(loadData, currentDefinition, delegate)
        }


        fileList.each {
            String totalLoadedRecordsSql = "select count(*) as total from ${getLoadingTable()} where source_file = '${it}'"
            int totalLoadingRecords = getSqlPlus().firstRow(totalLoadedRecordsSql).total
            String totalValidationErrorsSql = "select count(*) as total from ${validationErrorTable} where file_name = '${it}'"
            int totalValidationErrors = getSqlPlus().firstRow(totalValidationErrorsSql).total

            log.info("loaded ${totalLoadingRecords} records from ${it}")
            log.info("there were ${totalValidationErrors} validation errors for ${it}")
            int totalLines = totalLoadingRecords + totalValidationErrors
            assert lineCount.get() == totalLines: "There are ${lineCount.get()} lines in ${it}, but only ${totalLines} where processed"
            log.info("file line count and lines processed match")
        }


    }

    protected static ProcessorDefinition runClosure(Closure closure, ProcessorDefinition definition, RouteBuilder builder) {
        if (closure.getParameterTypes().size() == 2) {
            return closure.call(definition, builder)
        } else {
            return closure.call(definition)
        }
    }
}

@Slf4j
class CurrentProgressLogger {
    Map<String, Long> previousTime = [:]
    Map<String, Integer> operationCount = [:]
    private static CurrentProgressLogger currentProgressLogger

    private CurrentProgressLogger() {}

    static synchronized CurrentProgressLogger getInstance() {
        if (currentProgressLogger) {
            return currentProgressLogger
        }

        currentProgressLogger = new CurrentProgressLogger()
    }

    synchronized void update(String operation, int operations, int interval) {
        if (!previousTime.containsKey(operation)) {
            previousTime.put(operation, new Date().getTime())
            operationCount.put(operation, operations)
        } else {
            int previousCount = operationCount.get(operation)
            operationCount.put(operation, previousCount + operations)
        }

        def currentCount = operationCount.get(operation)
        if (currentCount >= interval) {
            double previousSeconds = TimeUnit.SECONDS.convert(previousTime.remove(operation), TimeUnit.MILLISECONDS)
            double currentSeconds = TimeUnit.SECONDS.convert(new Date().getTime(), TimeUnit.MILLISECONDS)
            int totalOperations = operations + operationCount.remove(operation)
            double operationsPerSecond = (totalOperations as Double) / (currentSeconds - previousSeconds) * 100
            operationsPerSecond = Math.round(operationsPerSecond);
            operationsPerSecond = operationsPerSecond / 100;
            log.info("PROCESSED: operation: [${operation}], operations: ${totalOperations}, operationsPerSecond: ${operationsPerSecond}")
        }
    }
}

class FileFilter implements GenericFileFilter {
    String fileNameFilter
    Closure fileFilter

    boolean accept(GenericFile file) {
        boolean nameOk = true
        boolean fileOk = true

        if (fileNameFilter != null) {
            nameOk = file.fileNameOnly =~ fileNameFilter
        }

        if (fileFilter != null) {
            fileOk = fileFilter.call(file)
        }

        return nameOk && fileOk
    }
}

class IteratorWrapper implements Iterator {

    Iterator iterator
    int offset = 0
    int limit = Integer.MAX_VALUE
    int line = 0


    boolean hasNext() {
        if (line < offset) {
            getToOffset()
        }

        if (line >= limit + offset) {
            return false
        }
        return iterator.hasNext()
    }

    Object next() {
        if (line < offset) {
            getToOffset()
        }
        if (line >= limit + offset) {
            throw new NoSuchElementException()
        }
        line++
        iterator.next()
    }

    void remove() {
        throw new UnsupportedOperationException("not supported")
    }

    private void getToOffset() {
        if (offset > 1) {
            (1..offset).each {
                if (iterator.hasNext()) {
                    iterator.next()
                }
                line++
            }
        }
    }
}
