/*
 * Copyright 2010 Trustees of the University of Pennsylvania Licensed under the
 * Educational Community License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License. You may
 * obtain a copy of the License at
 *
 * http://www.osedu.org/licenses/ECL-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS"
 * BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */
package metridoc.workflows.builder

import groovy.util.logging.Slf4j
import java.beans.PropertyDescriptor
import metridoc.utils.ClassUtils
import org.codehaus.groovy.grails.commons.spring.BeanConfiguration
import org.codehaus.groovy.grails.commons.spring.DefaultRuntimeSpringConfiguration
import org.codehaus.groovy.grails.commons.spring.RuntimeSpringConfiguration
import org.codehaus.groovy.runtime.InvokerHelper
import org.springframework.beans.BeanWrapper
import org.springframework.beans.PropertyValues
import org.springframework.beans.factory.NoSuchBeanDefinitionException
import org.springframework.beans.factory.support.AbstractBeanDefinition
import org.springframework.beans.factory.support.DefaultListableBeanFactory
import org.springframework.context.ApplicationContext
import org.springframework.core.io.Resource
import org.springframework.core.io.support.PathMatchingResourcePatternResolver
import org.springframework.util.StringUtils

/**
 * Created by IntelliJ IDEA.
 * User: tbarker
 * Date: 2/16/12
 * Time: 10:28 AM
 */
class MetridocBeanBuilder implements ConfigBuilder {

    public static final String METRIDOC_EXTENDABLE_FACTORY = "metridocExtendableFactory"
    protected static final String TYPE_MUST_BE_SPECIFIED = "at very least type must be specified in a bean declaration"
    protected static final String TYPE_MUST_NOT_BE_NULL_AND_MUST_BE_A_CLASS = "the type in a bean declaration must not be null and must be a class"

    RuntimeSpringConfiguration springConfig = new DefaultRuntimeSpringConfiguration(beanFactory: new MetridocBeanFactory())
    ApplicationContext currentContext
    boolean contextHasBeenBuilt = false
    ClassLoader classLoader = ClassUtils.defaultClassLoader

    private Closure processBeans = {String name, args ->
        def type = validateType(args)
        def constructorArgs = []
        def closure = null

        def hasConstructorArgumentsAndOrClosure = args.size() > 1
        if (hasConstructorArgumentsAndOrClosure) {
            (constructorArgs, closure) = createBeanConfigArgs(args, constructorArgs, closure)
        }

        addBeanConfigToSpringConfig(name, type, constructorArgs, closure)
    }

    protected Closure importBeans = {List items ->
        loadBeans(items)
    }

    protected Closure importModule = {Class<Script> script ->
        loadBeans(script)
    }

    private addBeanConfigToSpringConfig(String name, Class type, List constructorArgs, closure) {
        def beanConfig = springConfig.addSingletonBean(name, type, constructorArgs as List)

        if (closure) {
            new BeanHelper(springConfig: springConfig, beanConfig: beanConfig, builder: this).run closure
        } else {
            beanConfig.autowire = BeanConfiguration.AUTOWIRE_BY_NAME
        }
    }

    private List createBeanConfigArgs(args, List constructorArgs, closure) {
        def hasClosure = args[args.size() - 1] instanceof Closure
        if (hasClosure) {
            (constructorArgs, closure) = processWithPropertyProcessing(args, constructorArgs, closure)
        } else {
            constructorArgs = processWithoutPropertyProcessing(args, constructorArgs)
        }
        return [constructorArgs, closure]
    }

    MetridocBeanBuilder loadBeans(Closure closure) {
        loadBeans(new WrappedClosure(closure: closure))
        return this
    }

    MetridocBeanBuilder loadBeans(List beanFiles) {
        beanFiles.each {
            loadBeans(it)
        }

        return this
    }

    MetridocBeanBuilder loadBeans(String file) {
        loadBeans(new PathMatchingResourcePatternResolver().getResources(file))
        return this
    }

    MetridocBeanBuilder loadBeans(Resource[] resources) {
        for (Resource resource: resources) {
            GroovyShell shell = classLoader != null ? new GroovyShell(classLoader) : new GroovyShell();
            def script = shell.parse(new InputStreamReader(resource.getInputStream()), resource.filename);
            loadBeans(script)
        }

        return this
    }

    MetridocBeanBuilder loadBeans(Class<Script> script) {
        loadBeans(script.newInstance())
        return this
    }

    MetridocBeanBuilder loadBeans(Script script) {
        def factoryBeanConfig = springConfig.addSingletonBean(METRIDOC_EXTENDABLE_FACTORY, ExtendableFactory, [])
        factoryBeanConfig.factoryMethod = "instance"
        script.setBinding(new SpringBinding(builder: this))
        script.metaClass.methodMissing = processBeans
        script.metaClass.importBeans = importBeans
        script.metaClass.importModule = importModule

        script.run()
        return this
    }

    private List<Object> processWithoutPropertyProcessing(args, List constructorArgs) {
        int size = args.size()
        if (size > 2) {
            Range range = [1..<size]
            constructorArgs = (args as List).getAt(range)
        } else {
            constructorArgs = [args[1]]
        }
        return constructorArgs
    }

    private List processWithPropertyProcessing(args, List constructorArgs, closure) {
        int closureIndex = args.size() - 1
        closure = args[closureIndex]
        if (args.size() > 2) {
            def range
            if (closureIndex == 2) {
                range = 1
            } else {
                range = [1..<closureIndex]
            }

            range.each {
                constructorArgs.add(args[it])
            }
        }
        return [constructorArgs, closure]
    }

    private validateType(args) {
        assert args: TYPE_MUST_BE_SPECIFIED
        def type = args[0]
        assert type && type instanceof Class: TYPE_MUST_NOT_BE_NULL_AND_MUST_BE_A_CLASS
        return type
    }

    ApplicationContext createApplicationContext() {
        springConfig.applicationContext
    }

    void load(List modules) {
        loadBeans(modules)
    }

    void load(Object workflow) {
        loadBeans(workflow)
    }

    List getPipeline(String pipelineName) {
        return getCurrentContext().getBean(pipelineName, List)
    }

    def getBean(String name) {
        return getCurrentContext().getBean(name)
    }

    def <T> T getBean(String name, Class<T> clazz) {
        return getCurrentContext().getBean(name, clazz)
    }

    void configure(Closure closure) {
        loadBeans(closure)
    }

    ApplicationContext getCurrentContext() {

        if (currentContext) return currentContext
        contextHasBeenBuilt = true
        currentContext = createApplicationContext()
    }

    static void setVariableInExtendableFactory(String name, Object value, RuntimeSpringConfiguration springConfig) {
        def beanConfig = springConfig.addSingletonBean(name, value.getClass(), [name])
        ExtendableFactory.instance()."$name" = value
        beanConfig.factoryBean = MetridocBeanBuilder.METRIDOC_EXTENDABLE_FACTORY
        beanConfig.factoryMethod = "value"
    }

}

@Slf4j
class BeanHelper {
    MetridocBeanBuilder builder
    RuntimeSpringConfiguration springConfig
    BeanConfiguration beanConfig
    boolean autoConfig = true

    void run(Closure closure) {
        closure.resolveStrategy = Closure.DELEGATE_ONLY
        closure.delegate = this
        closure.call()
        if (autoConfig) {
            beanConfig.setAutowire(BeanConfiguration.AUTOWIRE_BY_NAME)
        }
    }

    def propertyMissing(String name) {

        def value

        if (builder.contextHasBeenBuilt) {
            value = builder.currentContext.getBean(name)
        } else if (springConfig.containsBean(name)) {
            value = springConfig.getBeanConfig(name).beanDefinition
        } else {
            MetridocBeanBuilder.setVariableInExtendableFactory(name, null, springConfig)
            value = springConfig.getBeanConfig(name).beanDefinition
        }

        return value
    }

    def propertyMissing(String name, value) {
        beanConfig.setPropertyValue(name, value)
    }


}

class SpringBinding extends Binding {
    public static final String APPLICATION_CONTEXT = "applicationContext"
    MetridocBeanBuilder builder
    Map variables

    @Override
    Map getVariables() {

        if (builder.contextHasBeenBuilt && variables) {
            return variables
        }

        if (builder.contextHasBeenBuilt) {
            variables = threadSafeGetVariables()
            return variables
        }

        return threadSafeGetVariables()
    }

    private synchronized Map threadSafeGetVariables() {
        def beanNames = builder.springConfig.getBeanNames()
        def result = [:]
        beanNames.each {
            result.put(it, getVariable(it))
        }

        return result
    }

    @Override
    Object getVariable(String name) {

        if (APPLICATION_CONTEXT == name) {
            return builder.springConfig.unrefreshedApplicationContext
        }

        def factory = ExtendableFactory.instance()
        def springConfig = builder.springConfig

        if (factory.properties.containsKey(name)) {
            return factory.value(name)
        }

        if (springConfig.containsBean(name)) {
            def applicationContext = springConfig.unrefreshedApplicationContext
            return new WrappedValue(applicationContext: applicationContext, name: name)
        }

        throw new MissingPropertyException(name, SpringBinding)
    }

    @Override
    void setVariable(String name, Object value) {
        MetridocBeanBuilder.setVariableInExtendableFactory(name, value, builder.springConfig)
    }
}

class ExtendableFactory {
    def properties = [:]
    private static ExtendableFactory instance

    private ExtendableFactory() {}

    static ExtendableFactory instance() {
        if (instance) return instance

        instance = new ExtendableFactory()
    }

    def value(String name) {
        properties[name]
    }

    def propertyMissing(String name, args) {
        properties[name] = args
    }
}

class WrappedClosure extends Script {
    Closure closure

    @Override
    Object run() {
        closure.setDelegate(this)
        closure.setResolveStrategy(Closure.DELEGATE_FIRST)
        closure.call()
    }
}

class WrappedValue {
    String name
    ApplicationContext applicationContext
    def bean

    def getBean() {
        if (bean) return bean

        try {
            bean = applicationContext.getBean(name)
        } catch (NoSuchBeanDefinitionException ex) {
            throw new MissingPropertyException(name, WrappedValue)
        }
    }

    def invokeMethod(String name, args) {
        return InvokerHelper.invokeMethod(getBean(), name, args)
    }

    def propertyMissing(name) {
        getBean()."$name"
    }

    def propertyMissing(name, value) {
        getBean()."$name" = value
    }

    def asType(Class clazz) {
        return applicationContext.getBean(name, clazz)
    }
}

class MetridocBeanFactory extends DefaultListableBeanFactory {
    protected String[] unsatisfiedNonSimpleProperties(AbstractBeanDefinition mbd, BeanWrapper bw) {
        Set<String> result = new TreeSet<String>();
        PropertyValues pvs = mbd.getPropertyValues();
        PropertyDescriptor[] pds = bw.getPropertyDescriptors();
        for (PropertyDescriptor pd: pds) {
            if (pd.getWriteMethod() != null && !isExcludedFromDependencyCheck(pd) && !pvs.contains(pd.getName())) {
                result.add(pd.getName());
            }
        }
        return StringUtils.toStringArray(result);
    }
}
